///////////////////////////////////////////////////////////////////////////////
// File:       BSTree.java
//
// Assignment: Assignment 4: Tree-Tree-Tree Unit
// Course:     CS367, Spring 2004
// Date:       4/30/04
//
// Student:    Justin Fiedler
// Lecture:    2:30 - 3:45 TR
// CS Login:   fiedler
///////////////////////////////////////////////////////////////////////////////


/**
 * Class that creates a tree ADT using array implementation.
 *
 * <p>Bugs: none
 *
 * @author Justin Fiedler
 */	
public class BSTree extends ABSTree
{   public BSTree()
    {	treeData = new String[8];
    }
	
    /**
     * Allows a new data item to be inserted into the tree
     * according to the specific tree implementation.
     *
     * @param item the data value to insert into the tree
     */
    public void insertItem(String item)
	{	int i = 1;
		while(treeData[i] != null)
		{	if(item.compareTo(treeData[i]) < 0)
			{	i = getLeftChildPos(i);
			}
			else if(item.compareTo(treeData[i]) > 0)
			{	i = getRightChildPos(i);
			}
			else
			{	break;
			}
			
			if(i >= treeData.length)
			{	expandTreeData();
			}
		}
		treeData[i] = item;	
	}
	
    /**
     * Returns the current number of items in the tree.
     * @return value returned is the number of items currently in the tree.
     */
    public int getItemCount()
	{	int count = 0;
		for(int i = 1; i < treeData.length; i++)
		{	if(treeData[i] != null) count++;
		}
		return count;
	}
    
    /**
     * This method returns the height of the tree.
	 * It should look through the tree and calculate the depth
     * of the lowest leaf in the tree and return that value.
	 * <p>The height of a one node tree is one.</p>
     *
     * @return the depth of the lowest leaf node in the tree
     */
    public int getHeight()
	{	int lastPosOnLevel = 0;
		int lowestPos      = 0;
		int height         = 0;
		
		for(int i = 1; i < treeData.length; i++)
		{	if(treeData[i] != null) lowestPos = i;
		}
		
		while(lowestPos > lastPosOnLevel)
		{	lastPosOnLevel = 2 * lastPosOnLevel + 1;
		}
		
		lastPosOnLevel += 1;
		while(lastPosOnLevel / 2 > 0)
		{	lastPosOnLevel /= 2;
			height++;
		}
		
		return height;
				
	}
	
	/**
	 * The funciton performs a binary search for the search term.
	 *
	 * @param search_term the string target value
	 * @return returns the string if found or null if not found
	 */
	public String search(String search_term)
	{	int i = getArrayPos(search_term);
		String s = null;
		if(i != 0)
			s = treeData[i];
		return s;
	}
	
	/**
	 * The funciton performs a binary search for the search term 
	 * returns the int pos of the 's' in the array 0 for not found
	 *
	 * @param s the string target value to search for
	 * @return returns the int pos of the 's' in the array 0 for not found
	 */
	private int getArrayPos(String s)
	{	int i = 0;
		boolean found = false;
		
		if(!isEmpty()) 
		{	i++;
			while(i < treeData.length && treeData[i] != null
			      && !found)
			{	
				if(s.compareTo(treeData[i]) < 0)
				{	i = getLeftChildPos(i);
				}
				else if(s.compareTo(treeData[i]) > 0)
				{	i = getRightChildPos(i);
				}
				else
				{	found = true;
				}
			}
			if(!found) i = 0;
		}
		return i;
	}
	
	/**
     * This method returns the array pos of the left child
     * of the node at position i. Note it does NOT check if this position
     * is valid or not.
	 * 
	 * @param i the position to find left child for
     * @return the array pos of the left child of item at position i in array
     */
	private int getLeftChildPos(int i)
	{	return 2 * i;
	}
	
	/**
     * This method returns the array pos of the right child
     * of the node at position i. Note it does NOT check if this position
     * is valid or not.
	 * 
	 * @param i the position to find right child for
     * @return the array pos of the right child of item at position i in array
     */
	private int getRightChildPos(int i)
	{	return 2 * i + 1;
	}
	
	/**
	 * Deletes the item identified by the parameter. In doing so, this
	 * function maintain the order, location and/or balance of the tree's
	 * nodes (if necessary).
	 * <p> if the delete operation fails for any reason the function
	 * should return null </p>
	 *
	 * @param item_to_delete The string value of the item to delete
	 * @return the return value is either the value of the string deleted
	 *         or <i>null</i> if the item could not be deleted.
	 *
	 */
	public String deleteItem(String item_to_delete)
	{	int i = getArrayPos(item_to_delete);
		String s = treeData[i];
		
		if(i != 0)
		{	int replace = 0;
			int lcp = getLeftChildPos(i);
			int rcp = getRightChildPos(i);
			String temp = null;
			
			if(lcp < treeData.length && treeData[lcp] != null)
			{	replace = getRightmostChildPos(lcp);
			}
			else if(rcp < treeData.length && treeData[rcp] != null)
			{	replace = getLeftmostChildPos(rcp);
			}
			
			temp = treeData[replace];
			if(replace != 0) deleteItem(treeData[replace]);
			treeData[i] = temp;	
		}
		
		return s;
	}
	
	/**
     * This method returns the array pos of the left most child
     * of the node at position 'root' or 'root' if no left child exists
     *  
	 * @param root of the subtree to search
     * @return the array pos of the left most child in subtree with 'root' root
     */
	private int getLeftmostChildPos(int root)
	{	int lcp = getLeftChildPos(root);
		
		if(lcp >= treeData.length)
		{	return root;
		}
		else if(treeData[lcp] == null)
		{	return root;
		}
		else
		{	return getLeftmostChildPos(lcp);
		}
	}
	
	/**
     * This method returns the array pos of the right most child
     * of the node at position 'root' or 'root' if no right child exists
     *  
	 * @param root of the subtree to search
     * @return the array pos of the right most child in subtree with 'root' root
     */
	private int getRightmostChildPos(int root)
	{	int rcp = getRightChildPos(root);
		
		if(rcp >= treeData.length)
		{	return root;
		}
		else if(treeData[rcp] == null)
		{	return root;
		}
		else
		{	return getRightmostChildPos(rcp);
		}
	}
	
	/**
	 * Traversal function returns the representation of the tree
	 * based on an in-order tree traversal: visit in order Left-Root-Right
	 *
	 * @return a string representation of the in-order traversal
	 */
	public String inorderTraversal()
	{	return inorderR(1);	
	}
	
	/**
	 * Traversal function returns the representation of the tree
	 * based on an in-order tree traversal: visit in order Left-Root-Right
	 *
	 * @param i the root position of the tree
	 * @return the string array representation of the in-order traversal
	 */
	private String inorderR(int i)
	{	String s = "";
		if(i < treeData.length && treeData[i] != null)
		{	s += inorderR(getLeftChildPos(i));
			s += treeData[i];
			s += " ";
			s += inorderR(getRightChildPos(i));
		}
		return s;
	}
		
	/**
	 * Traversal function returns the representation of the tree
	 * based on an pre-order tree traversal: visit in order Root-Left-Right
	 *
	 * @return the string array representation of the pre-order traversal
	 *
	 */
	public String preorderTraversal()
	{	return preorderR(1);
	}
	
	/**
	 * Traversal function returns the representation of the tree
	 * based on an pre-order tree traversal: visit in order Root-Left-Right
	 *
	 * @param i the root position of the tree
	 * @return the string array representation of the pre-order traversal
	 */
	private String preorderR(int i)
	{	String s = "";
		if(i < treeData.length && treeData[i] != null)
		{	s += treeData[i];
			s += " ";
			s += preorderR(getLeftChildPos(i));
			s += preorderR(getRightChildPos(i));
			
		}
		return s;
	}
	
	/**
	 * Traversal function returns the representation of the tree
	 * based on an post-order tree traversal: visit in order Left-Right-Root
	 *
	 * @return the string array representation of the post-order traversal
	 */
	public String postorderTraversal()
	{	return postorderR(1);
	}
	
	/**
	 * Traversal function returns the representation of the tree
	 * based on an post-order tree traversal: visit in order Left-Right-Root
	 *
	 * @param i the root position of the tree
	 * @return the string array representation of the post-order traversal
	 */
	private String postorderR(int i)
	{	String s = "";
		if(i < treeData.length && treeData[i] != null)
		{	s += postorderR(getLeftChildPos(i));
			s += postorderR(getRightChildPos(i));
			s += treeData[i];
			s += " ";
		}
		return s;
	}
	
	/**
	 *	Expands the tree array to twice its original size.
	 */
	private void expandTreeData()
	{	String[] temp = new String[2 * treeData.length];
		for(int i = 0; i < treeData.length; i++)
		{	temp[i] = treeData[i];
		}
		treeData = temp;
	}
	
	/**
	 * Returns true if tree is empty false if it is not
	 *
	 * @return boolean whether tree is empty or not
	 */
	private boolean isEmpty()
	{	return (treeData[1] == null);
	}
}
