///////////////////////////////////////////////////////////////////////////////
// File:       TestBSTree.java
//
// Assignment: Assignment 4: Tree-Tree-Tree Unit
// Course:     CS367, Spring 2004
// Date:       4/30/04
//
// Student:    Justin Fiedler
// Lecture:    2:30 - 3:45 TR
// CS Login:   fiedler
///////////////////////////////////////////////////////////////////////////////

import junit.framework.TestCase;
import junit.textui.TestRunner;

/**
 *  JUnit tests of the BSTree class.
 *  
 */
public class TestBSTree extends TestCase
{	
	String[] treeAfterDeleteH = {null, "F",
								 "B", "G",
								 "A", "D", null, null,
								 null, null, "C", "E", null, null, null, null};
	
	String[] treeAfterDeleteG = {null, "F",
								 "B", "H",
								 "A", "D", null, null,
								 null, null, "C", "E", null, null, null, null};
	
	String[] treeAfterDeleteED = {null, "F",
								 "B", "G",
								 "A", "C", null, "H"};
								 
	String[] treeAfterDeleteD = {null, "F",
								 "B", "G",
								 "A", "C", null, "H",
								 null, null, null, "E", null, null, null, null};
	
	String[] treeAfterDeleteF = {null, "E",
								 "B", "G",
								 "A", "D", null, "H",
								 null, null, "C", null, null, null, null, null};
	
	String[] tree2AfterDeleteA = {null, "B",
								  null, "D",
								  null, null, "C", "E"};
	
	
	
	/** 
	 *  All tests require an instance of a class that inherits from
	 *  the abstract super class ABSTree.
	 *  The instance is created once in the "setUp" method.
	 */
	private ABSTree tree = null;

	/** 
	 *  This method is kind of like a constructor for JUnit test classes.
	 *  It is automatically run before each test method.
	 *  Anything that should be done before running each test, 
	 *  should be put here instead of in each test method.
	 */
	public void setUp()
	{	tree = new BSTree();
	}

	/**
	 *	Creates a new tree
	 *	           F
	 *	       B       G
	 *	     A   D   _   H
	 *	    _ _ C E _ _ _ _
	 */ 
	public void createTestTree()
	{	setUp();
		tree.insertItem("F");
		tree.insertItem("B");
		tree.insertItem("A");
		tree.insertItem("D");
		tree.insertItem("C");
		tree.insertItem("E");
		tree.insertItem("G");
		tree.insertItem("H");		
	}
	
	/**
	 *	Creates a new tree
	 *	           A
	 *	       _       D
	 *	     _   _   B   E
	 *	    _ _ _ _ _ C _ _
	 */ 
	public void createTestTree2()
	{	setUp();
		tree.insertItem("A");
		tree.insertItem("D");
		tree.insertItem("B");
		tree.insertItem("C");
		tree.insertItem("E");		
	}
	
	/** Tests that the tree is empty when no items have been added. */
	public void testEmptyBSTree()
	{	assertEquals( "BSTree should be empty", 0, tree.getItemCount() ); 
	}
	
	public void testDeleteFromEmptyTree()
	{	try
		{	tree.deleteItem("1");
		}
		catch(Exception e)
		{	System.out.println("Error Removing From Empty Tree");
		}
	}
	
	public void testMultipleInsert()
	{	tree.insertItem("1");
		tree.insertItem("1");
		assertEquals( "Tree should only have 1 item", 1, tree.getItemCount() ); 
	}
	
	public void testHeight()
	{	assertEquals( "BSTree should have height 0", 0, tree.getHeight() );
		tree.insertItem("B");
		assertEquals( "BSTree should have height 1", 1, tree.getHeight() );
		tree.insertItem("C");
		assertEquals( "BSTree should have height 2", 2, tree.getHeight() );
		tree.insertItem("A");
		assertEquals( "BSTree should have height 2", 2, tree.getHeight() );
		tree.insertItem("D");
		assertEquals( "BSTree should have height 3", 3, tree.getHeight() );
	}
	
	public void testSearchItemAtRoot()
	{	createTestTree();
		assertEquals( "'F' should be found at root", "F", tree.search("F") );
	}
	
	public void testSearchItemInLeftSubTree()
	{	createTestTree();
		assertEquals( "'D' should be found in left subtree", "D", tree.search("D") );
	}
	
	public void testSearchItemInRightSubTree()
	{	createTestTree();
		assertEquals( "'H' should be found in right subtree", "H", tree.search("H") );
	}
	
	public void testSearchItemNotInTree()
	{	createTestTree();
		assertEquals( "Should return null for item not found", null, tree.search("4") );
	}
	
	public void testDeleteItemNotInTree()
	{	createTestTree();
		assertEquals( "Should return null for item not found", null, tree.deleteItem("4") );
		String[] treePrevious = tree.getArray();
		assertEquals( "Tree should not be mutated" , true, equalArray(treePrevious, tree.getArray()));
	}
	
	public void testDeleteItemWithNoChildren()
	{	createTestTree();
		assertEquals( "Should return 'H'", "H", tree.deleteItem("H") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(treeAfterDeleteH, tree.getArray()));
		assertEquals( "'H' should no longer appear in tree", null, tree.search("H") );
		assertEquals( "BSTree should have 7 items", 7, tree.getItemCount() );
		assertEquals( "Height should be 4", 4, tree.getHeight() );
	}
	
	public void testDeleteItemWithOnlyRightChildren()
	{	createTestTree();
		assertEquals( "Should return 'G'", "G", tree.deleteItem("G") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(treeAfterDeleteG, tree.getArray()));
		assertEquals( "'G' should no longer appear in tree", null, tree.search("G") );
		assertEquals( "BSTree should have 7 items", 7, tree.getItemCount() );
		assertEquals( "Height should be 4", 4, tree.getHeight() );
	}
	
	public void testDeleteItemWithOnlyLeftChildren()
	{	createTestTree();
		tree.deleteItem("E");
		assertEquals( "Should return 'D'", "D", tree.deleteItem("D") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(treeAfterDeleteED, tree.getArray()));
		assertEquals( "'D' should no longer appear in tree", null, tree.search("D") );
		assertEquals( "BSTree should have 6 items", 6, tree.getItemCount() );
		assertEquals( "Height should be 3", 3, tree.getHeight() );
	}
	
	public void testDeleteItemWithTwoChildren()
	{	createTestTree();
		assertEquals( "Should return 'D'", "D", tree.deleteItem("D") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(treeAfterDeleteD, tree.getArray()));
		assertEquals( "'D' should no longer appear in tree", null, tree.search("D") );
		assertEquals( "BSTree should have 7 items", 7, tree.getItemCount() );
		assertEquals( "Height should be 4", 4, tree.getHeight() );
	}
	
	public void testDeleteItemWithGrandChildren()
	{	createTestTree();
		assertEquals( "Should return 'F'", "F", tree.deleteItem("F") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(treeAfterDeleteF, tree.getArray()));
		assertEquals( "'F' should no longer appear in tree", null, tree.search("F") );
		assertEquals( "BSTree should have 7 items", 7, tree.getItemCount() );
		assertEquals( "Height should be 4", 4, tree.getHeight() );
	}
	
	public void testRecursiveDelete()
	{	createTestTree2();
		assertEquals( "Should return 'A'", "A", tree.deleteItem("A") );
		assertEquals( "BSTree array is not in proper order" , true, equalArray(tree2AfterDeleteA, tree.getArray()));
		assertEquals( "'A' should no longer appear in tree", null, tree.search("A") );
		assertEquals( "BSTree should have 4 items", 4, tree.getItemCount() );
		assertEquals( "Height should be 3", 3, tree.getHeight() );
	}
	
	public void testFreakishlyLargeTree()
	{	try
		{	String s = "";
			for(int i = 0; i < 20; i++)
			{	tree.insertItem(s + (char)('A' + i));
			}
		}
		catch(Exception e)
		{	System.out.println("BSTree should be able to handle a tree of any size");
		}
	}
	
	public void testInorderTraversal()
	{	createTestTree();
		assertEquals( "Inorder tranversal should be A B C D E F G H", "A B C D E F G H ", tree.inorderTraversal() );
	}
	
	public void testPreorderTraversal()
	{	createTestTree();
		assertEquals( "Pre-order tranversal should be F B A D C E G H", "F B A D C E G H ", tree.preorderTraversal() );
	}
	
	public void testPostorderTraversal()
	{	createTestTree();
		assertEquals( "Post-order tranversal should be A C E D B H G F", "A C E D B H G F ", tree.postorderTraversal() );
	}
	
	/**
	 * Tests if two string arrays are equal
	 *
	 * @return true if String[]'s are equal if not the returns false
	 */
	public boolean equalArray(String[] a, String [] b)
	{	int length1 = a.length < b.length ? a.length : b.length;
		int length2 = a.length > b.length ? a.length : b.length;
		
		boolean equal = true;
		
		for(int i = 0; i < length1; i++)
		{	if(a[i] != null && b[i] != null)
			{	if(a[i].compareTo(b[i]) != 0)
				{	equal = false;
				}
			}
			else if(a[i] != null || b[i] != null)
			{	equal = false;
			}
		}
		
		if(a.length > b.length)
		{	for(int i = length1; i < length2; i++)
			{	if(a[i] != null) equal = false;
			}
		}
		else
		{	for(int i = length1; i < length2; i++)
			{	if(b[i] != null) equal = false;
			}
		}		
		
		return equal;
	}
}
